// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       mgl_ui_event_default_delegate.cc
 *  \brief      MGL UIイベントのデフォルトデリゲート
 *  \date       Since: July 14, 2023. 16:29:27 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mgl/ui/mgl_ui_event_default_delegate.h>
#include <mgl/input/mouse/mgl_mouse.h>
#include <mgl/input/touch/mgl_touch.h>
#include <mgl/ui/mgl_ui_widget.h>

namespace MGL::UI
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      UIイベントの更新処理（外部呼び出し用）
 *  \param[in]  targets         イベント対象のUIウィジット
 *  \param[in]  currentFocusID  現在フォーカス中のウィジットのイベントID
 *  \return     イベント更新結果
 */
/* ------------------------------------------------------------------------- */
EventResult DefaultEventDelegate::Update(
        const MGL::STL::vector<EventTarget> &targets,
        EventIdentifier currentFocusID) noexcept
{
    auto result = UpdateMouseEvent(targets);
    if (result.type == EventType::None)
    {
        result = UpdateTouchEvent(targets, currentFocusID);
    }

    return result;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      UIイベントの更新処理
 *  \param[in]  targets         イベント対象のUIウィジット
 *  \param[in]  currentFocusID  現在フォーカス中のウィジットのイベントID
 *  \return     イベント更新結果
 */
/* ------------------------------------------------------------------------- */
EventResult DefaultEventDelegate::OnUpdateUIEvent(
        const MGL::STL::vector<EventTarget> &targets,
        EventIdentifier currentFocusID) noexcept
{
    return Update(targets, currentFocusID);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      マウス入力の更新処理
 *  \param[in]  targets     イベント対象のUIウィジット
 *  \return     イベント更新結果
 */
/* ------------------------------------------------------------------------- */
EventResult DefaultEventDelegate::UpdateMouseEvent(const MGL::STL::vector<EventTarget> &targets) noexcept
{
    MGL::Input::Mouse mouse;
    
    // マウスが接続されていなければ処理しない
    if (!mouse.IsConnecting())
    {
        return EventResult();
    }
    
    EventResult result;
    
    if (!mouse.GetDeltaMove().IsZero() || mouse.IsTriggered(MGL::Input::MouseButton::Left))
    {
        for (const auto &target : targets)
        {
            // マウスカーソルがウィジットの範囲内に入っていればフォーカス or 決定
            if (target.rectangle.IsEnclosed(mouse.GetPosition()))
            {
                // 左クリックされた場合は決定を送信（フォーカスも自動で行われる）
                if (mouse.IsTriggered(MGL::Input::MouseButton::Left))
                {
                    return EventResult(EventType::Decide, target.widget->GetEventIdentifier());
                }
                // 単にカーソルがウィジットの上に乗っているだけであればフォーカス
                else
                {
                    // 既にフォーカス中なら継続
                    if (target.widget->GetEventState() == EventState::Focusing)
                    {
                        return EventResult();
                    }

                    return EventResult(EventType::Focus, target.widget->GetEventIdentifier());
                }
            }
            // ウィジットがフォーカス中でカーソルが範囲外に出た場合はフォーカス解除
            else if (target.widget->GetEventState() == EventState::Focusing)
            {
                result = EventResult(EventType::Unfocus, target.widget->GetEventIdentifier());
            }
        }
    }
    
    return result;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      タッチ入力の更新処理
 *  \param[in]  targets     イベント対象のUIウィジット
 *  \return     イベント更新結果
 */
/* ------------------------------------------------------------------------- */
EventResult DefaultEventDelegate::UpdateTouchEvent(const MGL::STL::vector<EventTarget> &targets, EventIdentifier currentFocusID) noexcept
{
    // タッチデバイスが利用不可能な環境では処理しない
    if (!MGL::Input::Touch::IsAvailable())
    {
        return EventResult();
    }
    
    // フォーカスなし
    if (currentFocusID == kEventIDInvalid)
    {
        for (const auto &target : targets)
        {
            const auto *widget = target.widget;

            // 初回タッチ，または一度フォーカスしたタッチがウィジットの範囲内に入っている場合はフォーカスする
            if (MGL::Input::Touch touch; touch.IsEnabled())
            {
                if (touch.IsTouchedFirst() || (touch.GetTouchID() == widget->GetEventTouchID()))
                {
                    if (target.rectangle.IsEnclosed(touch.GetPosition()))
                    {
                        return EventResult(EventType::Focus, widget->GetEventIdentifier(), 0, touch.GetTouchID());
                    }
                }
            }
        }
    }
    // フォーカス中（ウィジットをタッチ中）
    else
    {
        for (const auto &target : targets)
        {
            const auto *widget = target.widget;
            
            // フォーカス中のウィジットのみを処理
            if (widget->GetEventIdentifier() == currentFocusID)
            {
                // タッチ継続中
                if (MGL::Input::Touch touch(widget->GetEventTouchID()); touch.IsEnabled())
                {
                    // タッチ位置がウィジットの範囲内であれば継続
                    if (target.rectangle.IsEnclosed(touch.GetPosition()))
                    {
                        return EventResult();
                    }
                    // タッチ位置が範囲外に出たらフォーカス解除（再度範囲内に入った場合に備えてタッチIDは伝えておく）
                    else
                    {
                        return EventResult(EventType::Unfocus, widget->GetEventIdentifier(), 0, touch.GetTouchID());
                    }
                }
                // フォーカス継続中にタッチが解除されたら決定
                else
                {
                    return EventResult(EventType::DecideAndUnfocus, widget->GetEventIdentifier(), 0, touch.GetTouchID());
                }
            }
        }
        
        // フォーカス中のウィジットが見つからなければフォーカス解除
        return EventResult(EventType::Unfocus, currentFocusID);
    }

    return EventResult();
}
}   // namespace MGL::UI

// vim: et ts=4 sw=4 sts=4
